//////////////////////////////////////////////////////////////////////////////////////
// File     : x16demo.c																//
// Function : based on linux demo application for Sensoray model 516/616 boards		//
// Author   : Jim Lamberson and Charlie Liu											//
// Revision :																		//
//            Sept, 2003   Jim Lamberson    release sx16-0.7						//
//            Oct., 2003   Charlie X. Liu   Added demo for using caption overlay	//
// Dean Anderson: updated linux demo for Windows driver
//																					//
//////////////////////////////////////////////////////////////////////////////////////

/*
Copyright (C) 2003  Sensoray Co., Inc.  This program is free software;
you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

You must obey the GNU General Public License in all respects for all of
the code used. If you modify this file, you may extend this exception
to your version of the file, but you are not obligated to do so. If you
do not wish to do so, delete this exception statement from your version.
*/

// The necessary headers /////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include "Windows.h"
#include <stdio.h>
#include "sx16ioctl.h"			    // ioctl definitions for sx16 driver.
#include "sx16core.h"           // x16 attributes.
#include <time.h>


// Manifest constants ////////////////////////////////////////////////////////////////////////////////

#define ENABLED				1
#define DISABLED			0

#define TRUE					1
#define FALSE					0

#define MAXLEN				80				// Maximum length of user command string.
#define MAXNUMBOARDS	4					// Maximum number of boards supported by the driver.

// Stream mode:
#define MODE_IDLE			0					// Halted.
#define MODE_PLAY			1					// Streaming from file.
#define MODE_RECORD		2					// Streaming to file.


// Configuration constants ///////////////////////////////////////////////////////////////////////////

#define NUM_BOARDS		4					// Number of x16 boards in the system (1 to 4).
#define MPEG_BUF_SIZE	(1024 * 31)			// Number of MPEG bytes to transfer in a single chunk.  Max size is 32k(must be multiple of 8, preferably 1024)
#define BLOCKING_MODE	TRUE			    // Set to TRUE to use blocking read/write operations.

// Device names for the x16 boards.
#define DEV_NAME_0		"\\\\.\\SX16WDM0"
#define DEV_NAME_1		"\\\\.\\SX16WDM1"
#define DEV_NAME_2		"\\\\.\\SX16WDM2"
#define DEV_NAME_3		"\\\\.\\SX16WDM3"


// Storage //////////////////////////////////////////////////////////////////

typedef struct BOARD {					// BOARD OBJECT -------
  HANDLE		handle;					//  Handle to x16 device.
  int			mode;						//  Stream mode.
  int			stop_rq;				//  Stream Halt requested.
  char			fname[MAXLEN];	//  File name for mpeg data.
  DWORD 		io_thread_id;		//  Handle to stream i/o thread.
  int           board_num;
  unsigned int	mpeg_count;			//  Number of MPEG bytes transferred.
  int			flags;					//  Status flags from i/o thread to main thread.
} BOARD;

static struct BOARD bd[NUM_BOARDS];		// Board objects.
static int					act_bdnum	= 0;		// Board 0 is active by default.


// Function prototypes ///////////////////////////////////////////////////////////////////////////////

static int app_init( void );
static void app_shutdown( void );
static void cmd_processor( void );
static void *io_thread_func( void* arg );
static void get_attributes( struct BOARD *b );
static void set_attribute( struct BOARD *b, int val, int minval, int maxval, int ioctl_num );
static void get_caption_settings( struct BOARD *b );
static void start_stream( struct BOARD *b, char *fname, int new_stream_mode );
static void stop_stream( struct BOARD *b );

static ULONG ioctl_get( HANDLE hdevice, DWORD code, ULONG arg)
{
    DWORD  numbytes;
    ULONG outbuf;
    DWORD outbufsize = sizeof(ULONG);
    LPVOID inbuf= (LPVOID) &arg;
    DWORD  inbufsize = sizeof (ULONG);
    if (DeviceIoControl( hdevice, code, inbuf, inbufsize, (LPVOID *) &outbuf, outbufsize, &numbytes, NULL))
        return outbuf;
    else
        return 0;
}


static BOOL ioctl_set( HANDLE hdevice, DWORD code, ULONG arg)
{
    DWORD  numbytes;
    void *inbuf= (void *) &arg;
    ULONG  inbufsize = sizeof (ULONG);
    BOOL res;
    res = DeviceIoControl( hdevice, code, (void *)inbuf, inbufsize, NULL, 0, &numbytes, NULL);
    return TRUE;
}

///////////////////////////////////////////////////////////////////////////
// Init the system and execute user commands until termination.

int main( int argc, char *argv[] )
{
  int rtnval;

  // Attempt to open and initialize all boards.
  rtnval = app_init();

  // If all boards successfully opened and initialized ...
  if ( rtnval == 0 )
  {
    // Execute user commands until Application Quit command is invoked.
    cmd_processor();
  }

  // Close all opened boards.
  app_shutdown();

  // Exit application.
  return 0;
}

///////////////////////////////////////////////////////////////////////
// Open and initialize one board.  Returns: 0=success, -1=error.

static int board_open( int board_num, char *dev_name )
{
  // Set pointer to board's object.
  struct BOARD *b = &bd[board_num];

  // Initialize object.
  b->mode		= MODE_IDLE;
  b->fname[0]	= 0;
  b->flags		= 0;
  b->board_num  = board_num;

  // Open device.

    b->handle = CreateFile(dev_name, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (b->handle == INVALID_HANDLE_VALUE)
	{
	    printf("Unable to open SX16WDM device %s - error %d\n", dev_name, GetLastError());
		return -1;
    }
   

  // Program the default settings.
  
  ioctl_set( b->handle, SX16_IOC_SET_COLORBARS,			DISABLED);
  ioctl_set( b->handle, SX16_IOC_SET_VIDEO_SYSTEM,	SYS_NTSC	);
  ioctl_set( b->handle, SX16_IOC_SET_BITRATE,				4000000		);
  ioctl_set( b->handle, SX16_IOC_SET_MPEG,				0	);
  ioctl_set( b->handle, SX16_IOC_SET_INPUT,					0					);
  ioctl_set( b->handle, SX16_IOC_SET_INPUT_TYPE,		CVIDEO		);
  ioctl_set( b->handle, SX16_IOC_SET_VBR,						DISABLED	);
  ioctl_set( b->handle, SX16_IOC_SET_AUDIO_VOLUME,	6					);
  ioctl_set( b->handle, SX16_IOC_SET_AUDIO_MUTE,		DISABLED	);

  // Set the caption defaults.
  ioctl_set( b->handle, SX16_IOC_SET_CAPTION_WINDOW,			1				);	// window #1
  ioctl_set( b->handle, SX16_IOC_SET_CAPTION_XSTART,			0				);	// 0
  ioctl_set( b->handle, SX16_IOC_SET_CAPTION_YSTART,			0				);	// 0
  ioctl_set( b->handle, SX16_IOC_SET_CAPTION_SCALE,				OVL_SCALE_X1	);	// x1
  ioctl_set( b->handle, SX16_IOC_SET_CAPTION_COLOR,				OVL_WHITE		);	// white
  ioctl_set( b->handle, SX16_IOC_SET_CAPTION_TRANSPARENCY,		OVL_TRANSP_T	);	// transparent
  ioctl_set( b->handle, SX16_IOC_SET_CAPTION_ON,				CAPTION_OFF		);	// off
  
  return 0;
}

///////////////////////////////////////////////////////////////////////
// Open and initialize all boards.  Returns: -1=error, 0=success.

static int app_init( void )
{
  int		i;
  char	dev_name[80];

  // For each board ...
  for ( i = 0; i < NUM_BOARDS; i++ )
  {
    // Specify board's device name.
    switch ( i )
    {
    case 0:	strcpy( dev_name, DEV_NAME_0 );	break;
    case 1:	strcpy( dev_name, DEV_NAME_1 );	break;
    case 2:	strcpy( dev_name, DEV_NAME_2 );	break;
    case 3:	strcpy( dev_name, DEV_NAME_3 );	break;
    }

    // Open device.  Return error if open failed.
    if ( board_open( i, dev_name ) == -1 ) {
        printf("could not open board %d\n", i);
        continue;
    }
  }

  return 0;
}

//////////////////////////////////////////////////////////////
// Terminate all stream threads and close all boards.

static void app_shutdown( void )
{
  struct BOARD *b;

  // For each board ...
  for ( b = bd; b < &bd[NUM_BOARDS]; b++ )
  {
    // We are finished if this board was never opened.
    if ( b->handle == INVALID_HANDLE_VALUE )
      break;

    // Terminate the board's I/O thread if it exists.
    stop_stream( b );

    // Close the board.
    CloseHandle( b->handle );
  }

  printf( "\n" );
}

//////////////////////////////////////////////////////////////
// Display the user prompt.

static void prompt( void )
{
  printf( "x16demo[%d]>", act_bdnum );
  fflush( stdout );
}

//////////////////////////////////////////////////////////////
// Display the command menu.

static void show_menu( void )
{
  printf( "+-SX16DEMO-COMMANDS------------------------------------------------------------+\n" );
  printf( "|   board n ....... select board: 1..4                                         |\n" );
  printf( "|   chan n ........ set video channel: 0 to 3                                  |\n" );
  printf( "|   get ........... display current settings                                   |\n" );
  printf( "|   intype n ...... set video type: 0=Composite, 1=S-Video                     |\n" );
  printf( "|   loopback ...... set video codec loopback (516 only): 0|1                   |\n" );
  printf( "|   rgpio n ....... display GPIO output states                                 |\n" );
  printf( "|   wgpio n ....... set GPIO output states: 0=off, 1=on                        |\n" );
  printf( "|   h[elp] ........ display this list                                          |\n" );
  printf( "|   q[uit] ........ quit program                                               |\n" );
  printf( "|   reset ......... reset board                                                |\n" );
  printf( "+-AUDIO------------------------------------------------------------------------+\n" );
  printf( "|   arate n ....... set bits/second:                                           |\n" );
  printf( "|                     1=32K, 2=48K, 3=56K, 4=64K, 5=80K, 6=96K, 7=112K, 8=128K |\n" );
  printf( "|                     9=160K, 10=192K, 11=224K, 12=256K, 13=320K, 14=384K      |\n" );
  printf( "|   aformat n ..... set format:                                                |\n" );
  printf( "|                     0=mpeg1/layer1, 1=mpeg1/layer2, 2=mp3, 3=ac3/2-ch        |\n" );
  printf( "|   stereo n ...... set stereo mode:                                           |\n" );
  printf( "|                     mpeg: 0=stereo, 1=joint, 2=2-ch, 3=1-ch                  |\n" );
  printf( "|                     ac3:  0=1-ch, 1=2-ch, 2=undef, 3=undef                   |\n" );
  printf( "|   mute n ........ [output]set mute enable: 0=no mute, 1=mute                 |\n" );
  printf( "|   vol n ......... [output]set volume: 0..63 (-dB units)                      |\n" );
  printf( "|   agcon n ....... [input]set audio agcon(0,1)                                |\n" );
  printf( "|   agcsens n ..... [input]set audio agc sensitivity (agcon=1)                 |\n" );
  printf( "|                           0=-3db, 1=0db, 2=+3db, 3=+9, 4=+15, 5=+21, 6=+27db |\n" );
  printf( "|   again n ....... [input]set audio amp gain(0-127)[-3dB-+60.5dB]             |\n" );
  printf( "|                          manual gain hen agcon=0 default value 6 = 0dB       |\n" );
  printf( "+-VIDEO------------------------------------------------------------------------+\n" );
  printf( "|   bars n ........ set output colorbars enable: 0=off, 1=on                   |\n" );
  printf( "|   bright n ...... set luminance brightness: 0..255                           |\n" );
  printf( "|   contrast n .... set luminance contrast: 0..255                             |\n" );
  printf( "|   vsat n ........ set chrominance saturation: 0..255                         |\n" );
  printf( "|   vsystem n ..... set video system: 1=PAL, 2=NTSC                            |\n" );
  printf( "|   vstatus ....... get status of video input                                  |\n" );
  printf( "+-MPEG-------------------------------------------------------------------------+\n" );
  printf( "|   gop n ......... set GOP: 0 to 31                                           |\n" );
  printf( "|   ipdist n ...... set IP distance: 0 to 3                                    |\n" );
  printf( "|   mpeg n ........ set mpeg type and picture size:                            |\n" );
  printf( "|                     0=mpeg2:D1-704, 1=mpeg2:2/3 D1, 2=mpeg2:1/2 D1,          |\n" );
  printf( "|                     3=mpeg1:SIF,    4=mpeg1:QSIF,                            |\n" );
  printf( "|                     5=mpeg2:D1-720, 6=mpeg2:D1-640                           |\n" );
  printf( "|   stream n ...... set stream type: 0=mux, 1=video es, 2=audio es             |\n" );
  printf( "|   avsel n ....... set mux a/v select: 0=aud+vid, 1=vid only, 2=aud only      |\n" );
  printf( "|   vrate n ....... set video bits/second: 100000..10000000                    |\n" );
  printf( "|   vbr n ......... set variable bit rate enable: 0=off, 1=on                  |\n" );
  printf( "+-STREAM-----------------------------------------------------------------------+\n" );
  printf( "|   play fname .... play from MPEG file; default name is \"tmp.mpg\"             |\n" );
  printf( "|   rec fname ..... record to MPEG file; default name is \"tmp.mpg\"             |\n" );
  printf( "|   stop .......... stop MPEG stream and close file                            |\n" );
  printf( "+-CAPTION----------------------------------------------------------------------+\n" );
  printf( "|   window n ...... set window number: 1 or 2                                  |\n" );
  printf( "|   xstart n ...... set window position for x_start: 8..700                    |\n" );
  printf( "|   ystart n ...... set window position for y_start: 0..478 / 574 (NTSC/PAL)   |\n" );
  printf( "|   scale n ....... set scale for font:                                        |\n" );
  printf( "|                     0=x1, 1=x2, 2=x2(only vertical), 3=x4                    |\n" );
  printf( "|   color n ....... set color for caption:                                     |\n" );
  printf( "|                     0=white, 1=black,                                        |\n" );
  printf( "|   transp n ...... set transparency attribute for caption:                    |\n" );
  printf( "|                     1=transparent,     2=black-background                    |\n" );
  printf( "|                     3=transp-text on white-bkgrnd                            |\n" );
  printf( "|                     4=black-text on white-bkgrnd                             |\n" );
  printf( "|   text string ... set text string for overlay:  any ascii string             |\n" );
  printf( "|   caption n ..... set caption off/on: 0=off, 1=on                            |\n" );
  printf( "+------------------------------------------------------------------------------+\n" );
}

///////////////////////////////////////////////////////////////////////////////
// Decode and execute user command.  Returns TRUE if Quit selected.

static int exec_command( char *cmd, char *arg )
{
  struct BOARD	*b;

  // Convert argument to integer value (not relevant for all commands).
  int arg_val = -1;
  sscanf( arg, "%d", &arg_val );

  // Decode and execute command -----------------------------------

  // Terminate application.
  if ( cmd[0] == 'q' )
    return TRUE;

  // Select board.
  if	( strcmp( cmd, "board" ) == 0 )
  {
    if ( ( arg_val >= 0 ) && ( arg_val <= NUM_BOARDS ) )
      act_bdnum = arg_val;
    else
      printf( "Illegal or missing argument; must be in range 0..%d\n", NUM_BOARDS );
    prompt();
    return FALSE;
  }

  // Show help.
  if ( ( cmd[0] == 'h' ) || ( cmd[0] == '?' ) )
  {
    show_menu();
    prompt();
    return FALSE;
  }

  // Prompt for user command.
  prompt();

  // Set pointer to active board's object.
  b = &bd[act_bdnum];

  // Set video attribute.
  if      ( strcmp( cmd, "bright"    ) == 0 )	set_attribute( b, arg_val, 0,	                255,        				SX16_IOC_SET_BRIGHTNESS    );
  else if ( strcmp( cmd, "intype"    ) == 0 )	set_attribute( b, arg_val, CVIDEO,	         	SVIDEO,    					SX16_IOC_SET_INPUT_TYPE    );
  else if ( strcmp( cmd, "chan"      ) == 0 )	set_attribute( b, arg_val, 0,					        3,          				SX16_IOC_SET_INPUT			   );
  else if ( strcmp( cmd, "bars"      ) == 0 )	set_attribute( b, arg_val, DISABLED,	        ENABLED,    				SX16_IOC_SET_COLORBARS		 );
  else if ( strcmp( cmd, "contrast"  ) == 0 )	set_attribute( b, arg_val, 0,	                255,		    				SX16_IOC_SET_CONTRAST      );
  else if	( strcmp( cmd, "vsystem"   ) == 0 )	set_attribute( b, arg_val, SYS_PAL,		        SYS_NTSC,   				SX16_IOC_SET_VIDEO_SYSTEM  );
  else if ( strcmp( cmd, "vstatus"   ) == 0 )
  {
    int sts = ioctl_get( b->handle, SX16_IOC_GET_INPUT_STATUS, 0 );
    printf( "sts=%x (%s)\n", sts, (sts & VIDEO_DET) == 0 ? "No video signal" : sts & F60HZ_DET ? "NTSC video detected" : "PAL video detected" );
  }
	// Set mpeg codec attribute.
	else if ( strcmp( cmd, "mpeg"      ) == 0 )	set_attribute( b, arg_val, SIZE704_2,	        SIZE640_2,  				SX16_IOC_SET_MPEG		       );
	else if ( strcmp( cmd, "stream"    ) == 0 )	set_attribute( b, arg_val, STREAMTYPE_MUX,    STREAMTYPE_TRANSPORT,			SX16_IOC_SET_STREAM_TYPE   );
	else if ( strcmp( cmd, "vrate"     ) == 0 )	set_attribute( b, arg_val, 100000,		        20000000,   				SX16_IOC_SET_BITRATE		   );
  else if ( strcmp( cmd, "vbr"       ) == 0 )	set_attribute( b, arg_val, DISABLED,	        ENABLED,    				SX16_IOC_SET_VBR			     );
  else if ( strcmp( cmd, "vsat"      ) == 0 )	set_attribute( b, arg_val, 0,	                255,		    				SX16_IOC_SET_SATURATION    );
  else if ( strcmp( cmd, "avsel"     ) == 0 )	set_attribute( b, arg_val, 0,	                2,	  		  				SX16_IOC_SET_AV_SELECT	   );
	else if ( strcmp( cmd, "prognum"   ) == 0 )	set_attribute( b, arg_val, 0,	                255,  		  				SX16_IOC_SET_PROG_NUM	     );  // not in menu
  else if ( strcmp( cmd, "gop"       ) == 0 )	set_attribute( b, arg_val, 0,	                31,	  	    				SX16_IOC_SET_M             );
  else if ( strcmp( cmd, "ipdist"    ) == 0 )	set_attribute( b, arg_val, 0,	                3,	  	    				SX16_IOC_SET_N             );

  // Set audio attribute.
  else if ( strcmp( cmd, "vol"       ) == 0 )	set_attribute( b, arg_val, 0,					63, 						SX16_IOC_SET_AUDIO_VOLUME  );
  else if ( strcmp( cmd, "mute"      ) == 0 )	set_attribute( b, arg_val, DISABLED,	        ENABLED,						SX16_IOC_SET_AUDIO_MUTE    );
  else if ( strcmp( cmd, "arate"     ) == 0 )	set_attribute( b, arg_val, 1,     		        14,        					SX16_IOC_SET_AUDIO_BITRATE );
  else if ( strcmp( cmd, "aformat"   ) == 0 )	set_attribute( b, arg_val, AFMT_MPEG1_LAYER1, AFMT_AC3,						SX16_IOC_SET_AUDIO_FORMAT  );
  else if ( strcmp( cmd, "asmprate"   ) == 0 )	set_attribute( b, arg_val, 0, 2,						SX16_IOC_SET_AUDIO_SMPRATE  );
  else if ( strcmp( cmd, "stereo"    ) == 0 )	set_attribute( b, arg_val, STEREOMODE_STEREO, STEREOMODE_SINGLE,	SX16_IOC_SET_AUDIO_STEREO  );
  else if ( strcmp( cmd, "agcon"     ) == 0 )	set_attribute( b, arg_val, 0,					1, 						    SX16_IOC_SET_AUDIO_AGCON  );
  else if ( strcmp( cmd, "agcsens"   ) == 0 )	set_attribute( b, arg_val, 0,					6, 						    SX16_IOC_SET_AUDIO_AGCSENS  );
  else if ( strcmp( cmd, "again"     ) == 0 )	set_attribute( b, arg_val, 0,					127,  					    SX16_IOC_SET_AUDIO_AMPGAIN  );
  else if ( strcmp( cmd, "treb"     ) == 0 )	set_attribute( b, arg_val, 0,					3, 						    SX16_IOC_SET_AUDIO_TREBLE  );
  else if ( strcmp( cmd, "bass"     ) == 0 )	set_attribute( b, arg_val, 0,					3, 						    SX16_IOC_SET_AUDIO_BASS  );

  // Set/get GPIO output states.
  else if ( strcmp( cmd, "wgpio"     ) == 0 ) ioctl_set( b->handle, SX16_IOC_SET_GPIO, arg_val );
  else if ( strcmp( cmd, "rgpio"     ) == 0 ) printf( "gpio=%d\n", ioctl_get( b->handle, SX16_IOC_GET_GPIO, arg_val ) );

	// Display current settings.
  else if ( strcmp( cmd, "get"       ) == 0 )	{ get_attributes( b ); get_caption_settings( b); prompt(); }

  // Control MPEG stream mode.
  else if ( strcmp( cmd, "rec"       ) == 0 )	start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_RECORD );
  else if ( strcmp( cmd, "play"      ) == 0 )	start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY   );
  else if ( strcmp( cmd, "stop"      ) == 0 )	stop_stream( b );

  // Set Caption attributes/parameters.
  else if ( strcmp( cmd, "window"    ) == 0 )	set_attribute( b, arg_val, 1,				2,					SX16_IOC_SET_CAPTION_WINDOW );
  else if ( strcmp( cmd, "xstart"    ) == 0 )	set_attribute( b, arg_val, 8,				MAX_XSTART,			SX16_IOC_SET_CAPTION_XSTART );
  else if ( strcmp( cmd, "ystart"    ) == 0 )	set_attribute( b, arg_val, 0,				MAX_YSTART_PAL,		SX16_IOC_SET_CAPTION_YSTART );
  else if ( strcmp( cmd, "scale"     ) == 0 )	set_attribute( b, arg_val, OVL_SCALE_X1,	OVL_SCALE_X4,		SX16_IOC_SET_CAPTION_SCALE );
  else if ( strcmp( cmd, "color"     ) == 0 )	set_attribute( b, arg_val, OVL_WHITE,		OVL_BLACK,			SX16_IOC_SET_CAPTION_COLOR );
  else if ( strcmp( cmd, "transp"    ) == 0 )	set_attribute( b, arg_val, OVL_TRANSP_T,	OVL_TRANSP_B_BKGRD,	SX16_IOC_SET_CAPTION_TRANSPARENCY );
  else if ( strcmp( cmd, "text"      ) == 0 )	ioctl_set( b->handle, SX16_IOC_SET_CAPTION_TEXT, (ULONG) arg );
  else if ( strcmp( cmd, "caption"   ) == 0 )	set_attribute( b, arg_val, CAPTION_OFF,		CAPTION_ON,			SX16_IOC_SET_CAPTION_ON );

	// Other commands.
	else if ( strcmp( cmd, "loopback"  ) == 0 ) ioctl_set( b->handle, SX16_IOC_GA_LOOPBACK, (ULONG) arg_val );
  else if ( strcmp( cmd, "reset"     ) == 0 ) ioctl_set( b->handle, SX16_IOC_RESET, 0 );


  // Illegal command.
  else if ( cmd[0] != 0 )
  {
    printf( "Invalid command: %s\n", cmd );
    prompt();
  }

  return FALSE;
}

////////////////////////////////////////////////////////////////////
// Execute user commands until Application Quit is requested.

static void cmd_processor( void )
{
  char					cmdbuf[MAXLEN];
  char					cmd[MAXLEN];
  char					arg[MAXLEN];
  char					*c;
//  int					arg_val;

  // Display the command menu.
  show_menu();

  // Prompt for user command.
  prompt();

  // Repeat until Application Terminate is requested ...
  for ( ; ; )
  {
    // Input user command string, and if no errors encountered ...
    if ( fgets( cmdbuf, MAXLEN - 1, stdin ) != NULL )
    {
      // Truncate newline char at end of command string.
      cmdbuf[ strlen( cmdbuf ) - 1 ] = 0;

      // If command line is not blank ...
      if ( cmdbuf[0] != 0 )
      {
        // Force command string to lower case, except command=="text".
		if ( strncmp( cmdbuf, "text", 4 ) != 0 )
			for ( c = cmdbuf; *c != 0; c++ ) 
				*c = tolower( *c );

        // Parse command string into opcode and argument.
        cmd[0] = arg[0] = 0;
        sscanf( cmdbuf, "%s %s", cmd, arg );

        // Decode and execute command
        if ( exec_command( cmd, arg ) )
          break;
      }
      else
      {
        // Prompt for user command.
        prompt();
      }
    }
  }
}

///////////////////////////////////////////////////
// Set programmable attribute on a board.

static void set_attribute( struct BOARD *b, int val, int minval, int maxval, int ioctl_num )
{
  // Set attribute if value is in valid range, otherwise display error message.
  if ( ( val >= minval ) && ( val <= maxval ) )
    ioctl_set( b->handle, ioctl_num, val );
  else
  {
    printf( "Illegal or missing argument; must be in range %d..%d\n", minval, maxval );
    prompt();
  }
}

/////////////////////////////////////////////////////
// Fetch and display board's current settings.

static char txt_streamtype	[][40] = { "0(mux)", "1(video-es)", "2(audio-es)", "3(transport)"};
static char txt_muxformat	[][40] = { "svcd", "vcd", "dvd", "cdda", "video-es", "audio-es", "transport" };
static char txt_arate		[][40] = {	"", "32", "48", "56", "64", "80", "96", "112", "128", "160", "192", "224", "256", "320", "384" };
static char txt_mpeg		[][40] = { "mpeg2:D1-704", "mpeg2:2/3-D1", "mpeg2:1/2-D1", "mpeg1:SIF", "mpeg1:QSIF", "mpeg2:D1-720", "mpeg2:D1-640" };
static char txt_boolean		[][40] = { "off", "on" };
static char txt_vsystem		[][40] = { "", "pal", "ntsc" };
static char txt_intype		[][40] = { "composite", "s-video" };
static char txt_sysformat	[][40] = { "svcd", "vcd", "other" };
static char txt_avsel		[][40] = { "audio+video", "video only", "audio only" };
static char txt_stereomode	[][40] = { "stereo", "joint stereo", "dual channel", "single channel" };
static char txt_asmprate	[][40] = { "44", "48", "32" };
static char txt_pcmres		[][40] = { "16", "18", "20", "24" };
static char txt_aformat		[][40] = { "mpeg1:layer-1", "mpeg1:layer-2", "mp3", "ac3:2-ch" };
static char txt_cap_scale	[][40] = { "x1", "x2", "x2 (vertical only)", "x4" };
static char txt_cap_color	[][40] = { "white", "black", "red", "green", "blue",  "yellow", "cyan", "magenta" };
static char txt_cap_transp	[][40] = { "", "transparent", "black-background", "transp-text on white-bkgrnd", "black-text on white-bkgrnd" }; 
static char txt_cap_on		[][40] = { "off", "on" };

// Fetch and display the attributes.
static void get_attributes( struct BOARD *b )
{
	printf( "\nCURRENT SETTINGS -----------\n" );
	printf( "audio bass           = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_bass					) );
	printf( "audio treble         = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_treble				) );
	printf( "audio volume         = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_volume				) );
	printf( "stereo mode          = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_stereo_mode	) );
	printf( "video bitrate (bps)  = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_bitrate			) );
	printf( "video brightness     = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_brightness		) );
	printf( "video channel        = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_chan					) );
	printf( "video contrast       = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_contrast			) );
	printf( "gop                  = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_m						) );
	printf( "ip dist              = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_n						) );
	printf( "video saturation     = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_saturation		) );
	printf( "program number       = %d\n", ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_program_number			) );

	printf( "audio format         = %s\n", txt_aformat		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_format						) ] );
	printf( "a/v select           = %s\n", txt_avsel		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_av_select							) ] );
	printf( "audio stereo mode    = %s\n", txt_stereomode	[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_stereo_mode			) ] );
	printf( "audio samprate (KHz) = %s\n", txt_asmprate		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_sample_rate			) ] );
	printf( "audio pcm res (bits) = %s\n", txt_pcmres		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_pcm_resolution		) ] );
	printf( "video type           = %s\n", txt_intype		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_type							) ] );
	printf( "audio bitrate (kbps) = %s\n", txt_arate		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_bitrate					) ] );
	printf( "video colorbars      = %s\n", txt_boolean		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_colorbars_enable	) ] );
	printf( "video filter         = %s\n", txt_boolean		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_filter_enable		) ] );
	printf( "video system         = %s\n", txt_vsystem		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_system						) ] );
	printf( "variable bitrate     = %s\n", txt_boolean		[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_vbr_enable				) ] );
	printf( "system format        = %s\n", txt_sysformat	[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_system_format		) ] );
	printf( "mpeg type/picsize    = %s\n", txt_mpeg			[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_video_picsize					) ] );
	printf( "mux format           = %s\n", txt_muxformat	[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_mux_format							) ] );
    printf( "stream type          = %s\n", txt_streamtype	[ ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_stream_type				) ] );
}

// Fetch and display the caption settings.
static void get_caption_settings( struct BOARD *b )
{
	printf( "caption window       = %d\n", ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_window			) );
	printf( "caption X start      = %d\n", ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_xstart			) );
	printf( "caption X stop       = %d\n", ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_xstop				) );
	printf( "caption Y start      = %d\n", ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_ystart			) );
	printf( "caption Y stop       = %d\n", ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_ystop				) );
	printf( "caption scale        = %s\n", txt_cap_scale	[ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_scale	) ] );
	printf( "caption color        = %s\n", txt_cap_color	[ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_color	) ] );
	printf( "caption transparency = %s\n", txt_cap_transp	[ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_transp	) ] );
	printf( "caption on/off       = %s\n", txt_cap_on		[ioctl_get( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_on		) ] );
}

///////////////////////////////////////////////////
// Start MPEG data streaming to/from a file.

static void start_stream( struct BOARD *b, char *fname, int new_stream_mode )
{
  HANDLE hThread;

    // If we are already streaming ...
  if ( b->mode != MODE_IDLE )
  {
    // Abort command.
    printf( "already streaming\n" );
    prompt();
    return;
  }

  if ( new_stream_mode == MODE_PLAY)
  {
    if ( ioctl_get( b->handle, SX16_IOC_GET_SETTING, ATTR_stream_type) == STREAMTYPE_TRANSPORT)
    { 
        printf( "Transport stream playback not currently supported.\n");
        prompt();
        printf("Change \"stream\" setting.\n");
        prompt();
        return;
    }
  }

  // Make the MPEG filename visible to the I/O thread.
  strcpy( b->fname, fname );

  // Indicate the new stream mode.
  b->mode = new_stream_mode;

  // Clear the stream thread's termination request.
  b->stop_rq = 0;

  // Create detached thread that will manage the MPEG stream flow.
//  pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
  hThread = CreateThread(NULL, 0, 
            (LPTHREAD_START_ROUTINE) io_thread_func, 
            b,  // b as argument 
            0, &b->io_thread_id); 
}


//////////////////////////////////////////////////
// Halt MPEG data stream to/from a file.

static void stop_stream( struct BOARD *b )
{
  // If stream i/o thread is running ...
  if ( b->mode != MODE_IDLE )
  {
    // Request thread termination.
    b->stop_rq = 1;

    // Wait for the thread to terminate
    while ( b->mode != MODE_IDLE )
      Sleep( 1 );
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////  MPEG STREAM I/O THREAD  //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////
// Copy a block of MPEG data from board to file.
// Returns number of bytes copied.

static int mpeg_to_file( struct BOARD *b, FILE *file )
{
  unsigned char	mpeg_buf[MPEG_BUF_SIZE];
  int		len;
  BOOL res;  
  // Attempt to read MPEG data from board.
  
  res = ReadFile(b->handle, mpeg_buf, MPEG_BUF_SIZE, &len, NULL);
  
  
  // Request Stream Halt and abort if we timed out while waiting to read MPEG data from board.
  if ( res == FALSE || len == 0)
  {
    printf( "No signal: Last Error %d\n", GetLastError() );
    prompt();
    b->stop_rq = 1;
    return 0;
  }
  else // Write MPEG data to file.
  {
	if (len <= MPEG_BUF_SIZE)
		fwrite( mpeg_buf, 1, len, file );
	else
		printf("invalid size\n");
  }

  // Return number of bytes transferred.
  return len;
}

/////////////////////////////////////////////////////
// Copy a block of MPEG data from file to board.
// Returns number of bytes copied.

static int mpeg_from_file( struct BOARD *b, FILE *file )
{
  char	mpeg_buf[MPEG_BUF_SIZE];
  int		len;
  int		remlen;
  int		nwritten;
  BOOL      bres;
  // Attempt to read MPEG data from file.
  len = fread( mpeg_buf, 1, MPEG_BUF_SIZE, file );

  // Request Stream Halt and abort if end of file reached.
  if ( len == 0 )
  {
    printf( "End of file reached\n" );
    prompt();
    b->stop_rq = 1;
    return 0;
  }

  // Repeat until all MPEG bytes have been transferred to the board ...
  for ( remlen = len; remlen > 0; )
  {
    // Attempt to write MPEG data to board.
    
    bres = WriteFile( b->handle, mpeg_buf, remlen, &nwritten, NULL );

    //  Request Stream Halt and abort if we timed out while writing MPEG data to board.
    if ( bres == FALSE || nwritten <=0 )
    {
      printf( "Player error: bytes written %d: Last error %d\n", nwritten, GetLastError() );
      prompt();
      b->stop_rq = 1;
      return 0;
    }

    // Compute remaining byte count.
    remlen -= nwritten;
  }

  // Return number of bytes transferred.
  return len;
}

////////////////////////////////////////////////////////////////////
// I/O THREAD MAIN.  Read/write MPEG data until stream is stopped.
// The arg is a pointer to the device's BOARD object.

void *io_thread_func( void *arg )
{
  FILE					*file;								// MPEG data file.
  unsigned int	rlen;									// MPEG bytes transferred.
  time_t				starttime;						// Stream start time.
  BOARD					*b = (BOARD *)arg;		// Cache pointer to board object.

  // Select the function that will be used to move MPEG data between file and board.
  int (* mpeg_transfer)( struct BOARD *b, FILE *file ) = ( b->mode == MODE_RECORD ) ? mpeg_to_file : mpeg_from_file;

  // Initialize stream statistics.
  b->mpeg_count = 0;		// No MPEG bytes have been transferred.
  time( &starttime );		// Log the start time.

  // Attempt to open the MPEG data file.  If successful ...
  if ( ( file = fopen( b->fname, ( b->mode == MODE_RECORD ) ? "wb" : "rb" ) ) != NULL )
  {
    // Report status.
    printf( "Streaming board %d, file %s\n", b->board_num, b->fname);
    prompt();
    // Until stream termination is requested ...
    while ( !b->stop_rq )
    {
      // Transfer a block of MPEG data between file and board.  If this is the first
      // transfer, the driver will automatically start the stream running.
      rlen = mpeg_transfer( b, file );

      // Collect statistical info.
      b->mpeg_count += rlen; 
      // Allow other processes to execute if we are in non-blocking mode and no data was available.
      if ( ( rlen == 0 ) && !BLOCKING_MODE )
      {
        // Sleep for approximately 10 milliseconds.
        Sleep( 10 );
      }
    }
    // Halt the stream.
    ioctl_set( b->handle, SX16_IOC_RW_STOP, 0 );
    // Close the MPEG data file.
    fclose( file );
    // Report statistics.
    printf( "%u MPEG bytes transferred, board %d\n", b->mpeg_count, b->board_num );
    prompt();
  }

  else
  {
    // Indicate there was a problem opening the MPEG data file.
    printf( "file open error\n" );
    prompt();
  }

  // Indicate stream is stopped.
  b->mode = MODE_IDLE;

  printf( "Stream stopped. board %d\n",b->board_num);
  prompt();

  // Terminate thread.
  return 0;
}

